@@ -2,6 +2,7 @@
 import logging
 import math
 import typing
+from collections import deque
 from dataclasses import dataclass
 from enum import Enum
 
@@ -55,12 +56,12 @@ class Template:
             [[base_mfcc[row][col]] for col in range(cols)] for row in range(rows)
         ]
 
-        dtw = DynamicTimeWarping()
+        avg_dtw = DynamicTimeWarping()
 
         # Collect features
         for template in templates[1:]:
-            dtw.compute_cost(template.mfcc, base_mfcc)
-            path = dtw.compute_path()
+            avg_dtw.compute_cost(template.mfcc, base_mfcc)
+            path = avg_dtw.compute_path()
             assert path is not None, "Failed to get DTW path"
             for row, col in path:
                 for i, feature in enumerate(template.mfcc[row]):
@@ -98,11 +99,13 @@ class Raven:
         templates: typing.List[Template],
         probability_threshold: typing.Tuple[float, float] = (0.45, 0.55),
         distance_threshold: float = 0.22,
-        dtw: typing.Optional[DynamicTimeWarping] = None,
+        frame_dtw: typing.Optional[DynamicTimeWarping] = None,
         dtw_window_size: int = 5,
+        dtw_step_pattern: float = 2,
         sample_rate: int = 16000,
         chunk_size: int = 960,
         shift_sec: float = 0.05,
+        before_chunks: int = 0,
         refractory_sec: float = 2.0,
         recorder: typing.Optional[WebRtcVadRecorder] = None,
         debug: bool = False,
@@ -116,6 +119,10 @@ class Raven:
         self.shift_sec = shift_sec
         self.sample_rate = sample_rate
 
+        self.before_buffer: typing.Optional[typing.Deque[bytes]] = None
+        if before_chunks > 0:
+            self.before_buffer = deque(maxlen=before_chunks)
+
         # Assume 16-bit samples
         self.sample_width = 2
         self.chunk_seconds = (self.chunk_size / self.sample_width) / self.sample_rate
@@ -124,8 +131,9 @@ class Raven:
         self.recorder = recorder or WebRtcVadRecorder()
 
         # Dynamic time warping calculation
-        self.dtw = dtw or DynamicTimeWarping()
+        self.dtw = frame_dtw or DynamicTimeWarping()
         self.dtw_window_size = dtw_window_size
+        self.dtw_step_pattern = dtw_step_pattern
 
         # Keep previously-computed distances and probabilities for debugging
         self.last_distances: typing.List[typing.Optional[float]] = [
@@ -161,7 +169,7 @@ class Raven:
         # State machine
         self.audio_buffer = bytes()
         self.state = RavenState.IN_SILENCE
-        self.num_silence_chunks = self.seconds_to_chunks(self.frame_duration_sec / 2)
+        self.num_silence_chunks = self.seconds_to_chunks(self.frame_duration_sec)
         self.silence_chunks_left = 0
         self.num_refractory_chunks = self.seconds_to_chunks(refractory_sec)
         self.refractory_chunks_left = 0
@@ -173,6 +181,12 @@ class Raven:
 
         Must be the same length as self.chunk_size.
 
+        Attributes
+        ----------
+
+        chunk: bytes
+          Raw audio chunk with one or more windows
+
         Returns
         -------
 
@@ -192,6 +206,14 @@ class Raven:
             self.state = RavenState.IN_SILENCE
             self.audio_buffer = bytes()
 
+            if self.before_buffer is not None:
+                self.before_buffer.clear()
+
+        # Keep chunks before detection
+        if self.before_buffer is not None:
+            self.before_buffer.append(chunk)
+
+        # Test chunk for silence/speech
         is_silence = self.recorder.is_silence(chunk)
 
         if self.state == RavenState.IN_SILENCE:
@@ -199,6 +221,14 @@ class Raven:
             if not is_silence:
                 # Start recording and checking audio
                 self.audio_buffer = bytes()
+
+                # Include some chunks before speech
+                if self.before_buffer is not None:
+                    for before_chunk in self.before_buffer:
+                        self.audio_buffer += before_chunk
+
+                    self.before_buffer.clear()
+
                 self.state = RavenState.IN_SPEECH
                 self.silence_chunks_left = self.num_silence_chunks
         else:
@@ -210,6 +240,9 @@ class Raven:
                 if self.silence_chunks_left <= 0:
                     # Back to sleep
                     self.state = RavenState.IN_SILENCE
+            else:
+                # Reset silence chunks
+                self.silence_chunks_left = self.num_silence_chunks
 
         if self.state == RavenState.IN_SPEECH:
             # Include audio for processing
@@ -232,6 +265,7 @@ class Raven:
                     self.refractory_chunks_left = self.num_refractory_chunks
                     return matching_indexes
 
+        # No detections
         return []
 
     def _process_frame(self, frame: np.ndarray) -> typing.List[int]:
@@ -246,30 +280,19 @@ class Raven:
         frame_mfcc = python_speech_features.mfcc(frame, self.sample_rate)
 
         for i, template in enumerate(self.templates):
-            # alignment = dtw(
-            #     template.mfcc,
-            #     frame_mfcc,
-            #     distance_only=True,
-            #     dist_method="cosine",
-            #     window_type="slantedband",
-            #     window_args={"window_size": self.dtw_window_size},
-            # )
-
-            # distance = alignment.distance / (len(frame_mfcc) + len(template.mfcc))
-
+            # Compute optimal distance with a window
             distance = self.dtw.compute_cost(
-                template.mfcc, frame_mfcc, self.dtw_window_size
+                template.mfcc,
+                frame_mfcc,
+                self.dtw_window_size,
+                step_pattern=self.dtw_step_pattern,
             )
 
+            # Normalize by sum of temporal dimensions
             normalized_distance = distance / (len(frame_mfcc) + len(template.mfcc))
 
-            probability = 1 / (
-                1
-                + math.exp(
-                    (normalized_distance - self.distance_threshold)
-                    / self.distance_threshold
-                )
-            )
+            # Compute detection probability
+            probability = self.distance_to_probability(normalized_distance)
 
             if self.debug:
                 _LOGGER.debug(
@@ -280,6 +303,7 @@ class Raven:
                     distance,
                 )
 
+            # Keep calculations results for debugging
             self.last_distances[i] = normalized_distance
             self.last_probabilities[i] = probability
 
@@ -288,6 +312,7 @@ class Raven:
                 < probability
                 < self.probability_threshold[1]
             ):
+                # Detection occured
                 matching_indexes.append(i)
 
         return matching_indexes
@@ -296,6 +321,16 @@ class Raven:
         """Compute number of chunks needed to cover some seconds of audio."""
         return int(math.ceil(seconds / (self.chunk_size / self.sample_rate)))
 
+    def distance_to_probability(self, normalized_distance: float) -> float:
+        """Compute detection probability using distance and threshold."""
+        return 1 / (
+            1
+            + math.exp(
+                (normalized_distance - self.distance_threshold)
+                / self.distance_threshold
+            )
+        )
+
     @staticmethod
     def wav_to_template(wav_file, name: str = "") -> Template:
         """Convert pre-trimmed WAV file to wakeword template."""
